

# From BLIP Diffusion
negative_prompt = "over-exposure, under-exposure, saturated, duplicate, out of frame, lowres, cropped, " \
                    "worst quality, low quality, jpeg artifacts, morbid, mutilated, out of frame, ugly, " \
                    "bad anatomy, bad proportions, deformed, blurry, duplicate"
